// Copyright (C) 2015-2016  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.

// Git-backup | Exception-style errors
package main

import (
    "fmt"
    "runtime"
    "strings"
)

// error type which is raised by raise(arg)
type Error struct {
    arg interface{}
    link *Error     // chain of linked Error(s) - see erraddcontext()
}

func (e *Error) Error() string {
    msgv := []string{}
    for e != nil {
        msgv = append(msgv, fmt.Sprintf("%v", e.arg))
        e = e.link
    }

    return strings.Join(msgv, ": ")
}

// raise error to upper level
func raise(arg interface{}) {
    // XXX is it ok?
    e, ok := arg.(*Error)
    if ok {
        panic(e)    // reraise
    }
    panic(&Error{arg, nil})
}

// raise formatted string
func raisef(format string, a ...interface{}) {
    raise(fmt.Sprintf(format, a...))
}


// raise if err != nil
// NOTE err can be != nil even if typed obj = nil:
//   var obj *T;
//   err = obj
//   err != nil     is true
func raiseif(err error) {
    //if err != nil && !reflect.ValueOf(err).IsNil() {
    if err != nil {
        raise(err)
    }
}


// checks recovered value to be of *Error
// if there is non-Error error - reraise it
// otherwise return Error either nil (no panic), or actual value
func _errcatch(r interface{}) *Error {
    e, _ := r.(*Error)
    //fmt.Printf("errcatch %#v  %#v\n", r, e)
    if e == nil && r != nil {
        panic(r)
    }
    return e
}

// catch error and call f(e) if it was caught.
// must be called under defer
func errcatch(f func (e *Error)) {
    e := _errcatch(recover())
    if e == nil {
        return
    }

    f(e)
}

// hook into unwinding process with f() call. Returned error is reraised.
// see also: erraddcontext()
// must be called under defer
func erronunwind(f func(e *Error) *Error) {
    // cannot do errcatch(...
    // as recover() works only in first-level called functions
    e := _errcatch(recover())
    if e == nil {
        return
    }

    e = f(e)
    panic(e)
}

/*
// add "prefix" context to error
func errlinkerror(e *Error, arg interface{}) *Error {
    return &Error{arg, e}
}
*/


// f is called if error unwinding is happening.
// call result is added to raised error as "prefix" context
func erraddcontext(f func() interface{}) {
    e := _errcatch(recover())
    if e == nil {
        return
    }

    arg := f()
    panic(&Error{arg, e})
}


// get name of currently running function (caller of myfuncname())
func myfuncname() string {
    pcv := [1]uintptr{}
    runtime.Callers(2, pcv[:])
    f := runtime.FuncForPC(pcv[0])
    if f == nil {
        return ""
    }
    return f.Name()
}

// add calling context to error
// n TODO
func erraddcallercontext(topfunc string, e *Error) *Error {
    // all callers
    var pcv = []uintptr{0}
    for {
        pcv = make([]uintptr, 2*len(pcv))
        n := runtime.Callers(1, pcv)
        if n < len(pcv) {
            pcv = pcv[:n]
            break
        }
    }

    // pcv -> function names
    funcv := []string{}
    frames := runtime.CallersFrames(pcv)
    for more := true; more; {
        var frame runtime.Frame
        frame, more = frames.Next()
        // do not go beyond topfunc
        if topfunc != "" && frame.Function == topfunc {
            break
        }
        // skip intermediates
        if strings.HasSuffix(frame.Function, "_") {
            continue
        }
        funcv = append(funcv, frame.Function)
    }

    // do not show anything after raise*()
    iraise := -1
    for i, funcname := range(funcv) {
        if strings.HasPrefix(funcname, "main.raise") {
            iraise = i
        }
    }
    funcv = funcv[iraise+1:]

    for _, f := range funcv {
        f = strings.TrimPrefix(f, "main.")
        e = &Error{f, e}
    }

    return e
}

// build error message associated with e
// should be called from under errcatch - in place which caught the error
func errmessage(e *Error) string {
    // all callers
    var pcv = []uintptr{0}
    for {
        pcv = make([]uintptr, 2*len(pcv))
        //n := runtime.Callers(0, pcv)
        n := runtime.Callers(1, pcv)
        if n < len(pcv) {
            pcv = pcv[:n]
            break
        }
    }

    // pcv -> function names
    funcv := []string{}
    frames := runtime.CallersFrames(pcv)
    for more := true; more; {
        var frame runtime.Frame
        frame, more = frames.Next()
//      // do not go beyond main
        if frame.Function == "main.main" {
        // do not go beyond caller
//      if len(funcv) > 0 && frame.Function == funcv[0] {
            break
        }
        // skip intermediates
        if strings.HasSuffix(frame.Function, "_") {
            continue
        }
        funcv = append(funcv, frame.Function)
    }

    // do not show anything after raise*()
    iraise := -1
    for i, funcname := range(funcv) {
        if strings.HasPrefix(funcname, "main.raise") {
            iraise = i
        }
    }
    funcv = funcv[iraise+1:]

    // print funcv prefix in top-down order
    msgprefix := []string{}
    for i := range(funcv) {
        f := funcv[len(funcv)-i-1]
        f = strings.TrimPrefix(f, "main.")
        msgprefix = append(msgprefix, f)
    }

    msg := strings.Join(append(msgprefix, ""), ": ")
    msg += fmt.Sprint(e)
    if !strings.HasSuffix(msg, "\n") {
        msg += "\n"
    }

    return msg
}
